"""Command line interface."""

import argparse
import sys
from collections.abc import Sequence
from pathlib import Path

DEFAULT_INIT = '''\
"""
Functional tests suite based on Turşu.

Documentation: https://mardiros.github.io/tursu/
"""
'''

DEFAULT_CONFTEST = """\
from tursu import tursu_collect_file

tursu_collect_file()
"""

DEFAULT_CONFTEST_WITH_DUMMIES = f'''\
import pytest

{DEFAULT_CONFTEST}

class DummyApp:
    """Represent a tested application."""
    def __init__(self):
        self.users = {{}}
        self.connected_user: str | None = None

    def login(self, username: str, password: str) -> None:
        if username in self.users and self.users[username] == password:
            self.connected_user = username


@pytest.fixture()
def app() -> DummyApp:
    return DummyApp()
'''

DEFAULT_STEPS = """\
from dataclasses import dataclass

from tursu import given, then, when

from .conftest import DummyApp


@dataclass
class User:
    username: str
    password: str


@given("a set of users:")
def a_set_of_users(app: DummyApp, data_table: list[User]):
    for user in data_table:
        app.users[user.username] = user.password


@when("{username} login with password {password}")
def login(app: DummyApp, username: str, password: str):
    app.login(username, password)


@then("the user is connected with username {username}")
def assert_connected(app: DummyApp, username: str):
    assert app.connected_user == username


@then("the user is not connected")
def assert_not_connected(app: DummyApp):
    assert app.connected_user is None

"""

DEFAULT_FEATURE = """\
Feature: User login with their own password

  Background:
    Given a set of users:
      | username | password      |
      | Bob      | dumbsecret    |
      | Alice    | anothersecret |

  Scenario: User can login
    When Bob login with password dumbsecret
    Then the user is connected with username Bob

  Scenario: User can't login with wrong password
    When Bob login with password notthat
    Then the user is not connected

  Scenario Outline: User can't login with someone else username
    When <username> login with password <password>
    Then the user is not connected

    Examples:
      | username | password      |
      | Bob      | anothersecret |
      | Alice    | dumbsecret    |

"""


def init(outdir: str, overwrite: bool, no_dummies: bool) -> None:
    """
    Will write a functional tests directory.

    :param outdir: destination
    :param overwrite: overwrite existing file.
    :param no_dummies: do not write dummy fixtures and tests.
    """
    with_dummies = not no_dummies
    outpath = Path(outdir)
    if outpath.exists() and not overwrite:
        print(f"{outdir} already exists")
        sys.exit(1)

    if outpath.is_file():
        outpath.unlink()

    outpath.mkdir(exist_ok=True, parents=True)
    (outpath / "__init__.py").write_text(DEFAULT_INIT)
    (outpath / "conftest.py").write_text(
        DEFAULT_CONFTEST_WITH_DUMMIES if with_dummies else DEFAULT_CONFTEST
    )

    if with_dummies:
        (outpath / "steps.py").write_text(DEFAULT_STEPS)
        (outpath / "login.feature").write_text(DEFAULT_FEATURE)


def main(args: Sequence[str] = sys.argv) -> None:
    """Entrypoint of the CLI."""
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(title="action", required=True)

    sp_action = subparsers.add_parser("init")
    sp_action.add_argument(
        "-o",
        "--out-dir",
        dest="outdir",
        default="tests/functionals",
        help="Directory where the handlers will be generated",
    )
    sp_action.add_argument(
        "--overwrite", action="store_true", dest="overwrite", default=False
    )
    sp_action.add_argument(
        "--no-dummies", action="store_true", dest="no_dummies", default=False
    )

    sp_action.set_defaults(handler=init)
    kwargs = parser.parse_args(args[1:])
    kwargs_dict = vars(kwargs)

    handler = kwargs_dict.pop("handler")
    handler(**kwargs_dict)



"""
Easily run Gherkin tests.
"""

from importlib import metadata

from tursu.entrypoints.plugin import tursu_collect_file
from tursu.runtime.pattern_matcher import RegEx
from tursu.runtime.registry import given, then, when

__version__ = metadata.version("tursu")

__all__ = [
    "given",
    "when",
    "then",
    "RegEx",
    "tursu_collect_file",
]



"""AST helpers at the test function level."""

import ast
import json
import re
from collections.abc import Mapping, Sequence
from typing import Annotated, Any, TypeGuard, cast, get_args, get_origin

from tursu.domain.model.gherkin import (
    GherkinExamples,
    GherkinFeature,
    GherkinKeyword,
    GherkinRule,
    GherkinScenario,
    GherkinScenarioOutline,
    GherkinStep,
)
from tursu.domain.model.steps import StepKeyword
from tursu.runtime.registry import Tursu


def repr_stack(stack: Sequence[Any]) -> Sequence[str]:
    """Helper to get a stack representation to display fancy tests results."""
    ret = []
    for el in stack:
        ret.append(repr(el))
    return ret


def is_step_keyword(value: GherkinKeyword) -> TypeGuard[StepKeyword]:
    """
    Typeguard for step keyword only.

    :param value: any gherkin keyword input
    """
    return value in get_args(StepKeyword)


def sanitize(name: str) -> str:
    """Used to generate valid python identifiers."""
    return re.sub(r"\W+", "_", name)[:100]


class TestFunctionWriter:
    """
    Helper to write a test function for a given scenario.

    :param scenario: scenario to compile to ast code.
    :param registry: tursu registry containing steps definition.
    :param steps: steps to include, including the one from the background.
    :param stack: tursu compilation current stack.
    :param package_name: package name where the scenario lives in.
    """

    def __init__(
        self,
        scenario: GherkinScenario | GherkinScenarioOutline,
        registry: Tursu,
        steps: Sequence[GherkinStep],
        stack: Sequence[Any],
        package_name: str,
    ) -> None:
        self.registry = registry
        self.gherkin_keyword: StepKeyword | None = None
        self.package_name = package_name

        self.fixtures = self.build_fixtures(steps, registry)
        decorator_list = self.build_tags_decorators(stack)
        examples_keys = None
        if isinstance(scenario, GherkinScenarioOutline) and scenario.examples:
            examples_keys = [c.value for c in scenario.examples[0].table_header.cells]
            params = ",".join(examples_keys)
            params_name = ast.Constant(params)
            data: list[ast.expr] = []
            for ex in scenario.examples:
                id_ = ex.name or ex.keyword
                for row in ex.table_body:
                    parametrized_set = ast.Attribute(
                        value=ast.Name(id="pytest", ctx=ast.Load()),
                        attr="param",
                        ctx=ast.Load(),
                    )
                    dataset: list[ast.expr] = [ast.Constant(c.value) for c in row.cells]
                    data.append(
                        ast.Call(
                            func=parametrized_set,
                            args=dataset,
                            keywords=[ast.keyword("id", ast.Constant(id_))],
                        )
                    )
            ex_args: list[ast.expr] = [
                params_name,
                ast.List(elts=data, ctx=ast.Load()),
            ]

            decorator = ast.Attribute(
                value=ast.Name(id="pytest", ctx=ast.Load()),
                attr="mark",
                ctx=ast.Load(),
            )

            parametrize_decorator = ast.Attribute(
                value=decorator, attr="parametrize", ctx=ast.Load()
            )

            decorator_list.append(
                ast.Call(func=parametrize_decorator, args=ex_args, keywords=[])
            )

        self.step_list: list[ast.stmt] = []
        runner_instance = ast.With(
            items=[
                ast.withitem(
                    context_expr=ast.Call(
                        func=ast.Name(id="TursuRunner", ctx=ast.Load()),
                        args=[
                            ast.Name(id="request", ctx=ast.Load()),
                            ast.Name(id="capsys", ctx=ast.Load()),
                            ast.Name(id="tursu", ctx=ast.Load()),
                            ast.Constant(value=repr_stack(stack)),
                        ],
                        keywords=[],
                    ),
                    optional_vars=ast.Name(id="tursu_runner", ctx=ast.Store()),
                )
            ],
            body=self.step_list,
            lineno=scenario.location.line + 2,
        )

        docstring = f"{scenario.name}\n\n    {scenario.description}".strip()

        args = self.build_args(self.fixtures, examples_keys)
        self.funcdef = ast.FunctionDef(
            name=f"test_{scenario.id}_{sanitize(scenario.name)}",
            args=ast.arguments(
                args=args,
                posonlyargs=[],
                kwonlyargs=[],
                kw_defaults=[],
                defaults=[],
            ),
            body=[
                ast.Expr(
                    value=ast.Constant(docstring),
                    lineno=scenario.location.line + 1,
                ),
                runner_instance,
            ],
            decorator_list=decorator_list,
            lineno=scenario.location.line,
        )

    def build_args(
        self, fixtures: Mapping[str, Any], examples_keys: Sequence[Any] | None = None
    ) -> list[ast.arg]:
        """Build the args for the test functions."""
        args = [
            ast.arg(
                arg="request",
                annotation=ast.Name(id="pytest.FixtureRequest", ctx=ast.Load()),
            ),
            ast.arg(
                arg="capsys",
                annotation=ast.Name(id="pytest.CaptureFixture[str]", ctx=ast.Load()),
            ),
            ast.arg(
                arg="tursu",
                annotation=ast.Name(id="Tursu", ctx=ast.Load()),
            ),
        ]
        for key, _val in fixtures.items():
            if key in ("request", "capsys", "tursu"):
                continue
            args.append(
                ast.arg(
                    arg=key,
                    annotation=ast.Name(id="Any", ctx=ast.Load()),
                )
            )
        if examples_keys:
            for exkeys in examples_keys:
                args.append(
                    ast.arg(
                        arg=exkeys,
                        annotation=ast.Name(id="str", ctx=ast.Load()),
                    )
                )
        return args

    def build_fixtures(
        self, steps: Sequence[GherkinStep], registry: Tursu
    ) -> dict[str, type]:
        """
        Get the fixture for the given step.

        :param steps: steps to include, including the one from the background.
        :param registry: tursu registry containing steps definition.
        :return: the fixtures to include for all the steps of the scenario.
        """
        fixtures: dict[str, type] = {}
        step_last_keyword = None
        for step in steps:
            if step.keyword_type == "Conjunction":
                if step_last_keyword is None:
                    raise ValueError(f'Using "{step.keyword}" keyword without context')
            else:
                step_last_keyword = step.keyword
            assert is_step_keyword(step_last_keyword)

            fixtures.update(
                registry.extract_fixtures(
                    self.package_name, step_last_keyword, step.text
                )
            )
        return fixtures

    def build_tags_decorators(self, stack: Sequence[Any]) -> list[ast.expr]:
        """
        Generate pytest markers for the function.
        It will traverse the stack to get all the gherkin tags in order to generate
        pytest markers with them.

        :param stack: current compiler stack.
        """
        decorator_list = []
        tags = self.get_tags(stack)
        if tags:
            for tag in tags:
                decorator = ast.Attribute(
                    value=ast.Name(id="pytest", ctx=ast.Load()),
                    attr="mark",
                    ctx=ast.Load(),
                )
                tag_decorator = ast.Attribute(value=decorator, attr=tag, ctx=ast.Load())
                decorator_list.append(tag_decorator)
        return decorator_list  # type: ignore

    def get_tags(self, stack: Sequence[Any]) -> set[str]:
        """
        Get the all the gherkin tags from the stack.

        :param stack: current compiler stack.
        """
        ret = set()
        for el in stack:
            match el:
                case (
                    GherkinFeature(tags=tags)
                    | GherkinRule(tags=tags)
                    | GherkinScenario(tags=tags)
                    | GherkinScenarioOutline(tags=tags)
                    | GherkinExamples(tags=tags)
                ):
                    for tag in tags:
                        ret.add(tag.name)
                case _:
                    ...
        return ret

    def get_keyword(self, stp: GherkinStep) -> StepKeyword:
        """
        Get the step keyword from the current step.
        If the step is a conjunction (`And`, `But`), the step keyword come from
        the previous step.

        :param stp: the step to analyse.
        :return: proper keyword for the step.
        :raises ValueError: if a conjunction has been used to start a scenario.
        """
        keyword = stp.keyword
        if stp.keyword_type == "Conjunction":
            if self.gherkin_keyword is None:
                raise ValueError(f'Using "{stp.keyword}" keyword without context')
            keyword = self.gherkin_keyword
        assert is_step_keyword(keyword)
        self.gherkin_keyword = keyword
        return keyword

    def build_step_args(
        self,
        step_keyword: StepKeyword,
        stp: GherkinStep,
        examples: Sequence[GherkinExamples] | None = None,
    ) -> list[ast.expr]:
        """
        Get a step ast argument.

        :param step_keyword: the step definition keyword.
        :param stp: the step to analyse.
        :param examples: in cases of scenario outline, its associated examples.
            In this case, the chevron-enclosed placeholders (e.g. `<value>`) are
            replaced by parametrized value during the call.
        :return: ast values for the step argument.
        :raises ValueError: if a conjunction has been used to start a scenario.
        """
        call_format_node = None
        text = ast.Constant(value=stp.text)
        if examples:
            format_keywords = []
            ex = examples[0]
            for cell in ex.table_header.cells:
                format_keywords.append(
                    ast.keyword(
                        arg=cell.value, value=ast.Name(id=cell.value, ctx=ast.Load())
                    )
                )
            call_format_node = ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id="tursu_runner", ctx=ast.Load()),
                    attr="format_example_step",
                    ctx=ast.Load(),
                ),  # tursu.run_step
                args=[
                    text,
                ],
                keywords=format_keywords,
            )
        return [
            ast.Constant(value=step_keyword),
            call_format_node if call_format_node else text,
        ]

    def parse_doc_string(
        self, step_keyword: StepKeyword, stp: GherkinStep
    ) -> ast.keyword:
        registry_step = self.registry.get_step(
            self.package_name, step_keyword, stp.text
        )

        assert registry_step, "Step not found"
        assert stp.doc_string, "Step has not doc_string"

        typ = None
        is_list = False
        anon = registry_step.pattern.signature.parameters["doc_string"].annotation
        if anon:
            param_origin = get_origin(anon)
            if param_origin is Annotated:
                # we are in a factory
                typ = get_args(anon)[-1]
            elif param_origin and issubclass(param_origin, Sequence):
                is_list = True
                typ = get_args(anon)[0]
                orig = get_origin(typ)
                if orig is dict:
                    typ = None
                elif orig is Annotated:
                    typ = get_args(typ)[-1]
            else:
                orig = get_origin(anon)
                if orig is dict:
                    typ = None
                else:
                    typ = anon

        if typ and typ is not str and stp.doc_string.media_type != "python":
            doc_string_keywords = []
            call_doc_string_node: ast.expr
            if isinstance(stp.doc_string.content, str):
                data = json.loads(stp.doc_string.content)
            else:
                data = stp.doc_string.content
            if is_list:
                doc_string_models: list[ast.expr] = []
                for model in cast(Sequence[Mapping[str, Any]], data):
                    doc_string_keywords = [
                        ast.keyword(
                            arg=key,
                            value=ast.Constant(value=val),
                        )
                        for key, val in model.items()
                    ]

                    param_origin = get_origin(anon)
                    doc_string_models.append(
                        ast.Call(
                            func=ast.Name(
                                id=self.registry.get_models_type(
                                    self.package_name, typ
                                ),
                                ctx=ast.Load(),
                            ),
                            keywords=doc_string_keywords,
                        )
                    )

                call_doc_string_node = ast.List(elts=doc_string_models, ctx=ast.Load())
            else:
                doc_string_keywords = [
                    ast.keyword(
                        arg=key,
                        value=ast.Constant(value=val),
                    )
                    for key, val in cast(Mapping[str, Any], data).items()
                ]

                param_origin = get_origin(anon)
                call_doc_string_node = ast.Call(
                    func=ast.Name(
                        id=self.registry.get_models_type(self.package_name, typ),
                        ctx=ast.Load(),
                    ),
                    keywords=doc_string_keywords,
                )
            return ast.keyword(arg="doc_string", value=call_doc_string_node)

        return ast.keyword(
            arg="doc_string", value=ast.Constant(value=stp.doc_string.content)
        )

    def parse_data_table(
        self, step_keyword: StepKeyword, stp: GherkinStep
    ) -> ast.keyword:
        """
        Parse the data table to its step definition signature.

        :param step_keyword: the step definition keyword.
        :param stp: the step to analyse.
            the step must have a definved data_table.
        """
        registry_step = self.registry.get_step(
            self.package_name, step_keyword, stp.text
        )

        assert registry_step, "Step not found"
        assert stp.data_table, "Step has no data_table"
        typ: type | None = None
        is_reversed = False
        anon = registry_step.pattern.signature.parameters["data_table"].annotation
        if anon:
            param_origin = get_origin(anon)
            if param_origin is Annotated:
                # we are in a factory
                is_reversed = True
                typ = get_args(anon)[-1]
            elif param_origin and issubclass(param_origin, Sequence):
                typ = get_args(anon)[0]
                orig = get_origin(typ)
                if orig is dict:
                    typ = None
                elif orig is Annotated:
                    typ = get_args(typ)[-1]
            else:
                is_reversed = True
                orig = get_origin(anon)
                if orig is dict:
                    typ = None
                else:
                    typ = anon

        if typ is None:
            if is_reversed:
                rev_tabl = {
                    row.cells[0].value: row.cells[1].value
                    for row in stp.data_table.rows
                }
                return ast.keyword(arg="data_table", value=ast.Constant(value=rev_tabl))
            else:
                tabl = []
                hdr = [c.value for c in stp.data_table.rows[0].cells]
                for row in stp.data_table.rows[1:]:
                    vals = [c.value for c in row.cells]
                    tabl.append(dict(zip(hdr, vals)))

                return ast.keyword(arg="data_table", value=ast.Constant(value=tabl))

        if is_reversed:
            datatable_keywords = [
                ast.keyword(
                    arg=row.cells[0].value,
                    value=ast.Constant(value=row.cells[1].value),
                )
                for row in stp.data_table.rows
                if row.cells[1].value != self.registry.DATA_TABLE_EMPTY_CELL
            ]
            call_rev_datatable_node = ast.Call(
                func=ast.Name(
                    id=self.registry.get_models_type(self.package_name, typ),
                    ctx=ast.Load(),
                ),
                keywords=datatable_keywords,
            )
            return ast.keyword(arg="data_table", value=call_rev_datatable_node)
        else:
            # we have to parse the value
            tabl = []
            hdr = [c.value for c in stp.data_table.rows[0].cells]
            call_datatable_node: list[ast.expr] = []
            for row in stp.data_table.rows[1:]:
                vals = [c.value for c in row.cells]
                datatable_keywords = []
                for key, val in zip(hdr, vals):
                    if val == self.registry.DATA_TABLE_EMPTY_CELL:
                        # empty string are our null value
                        continue
                    datatable_keywords.append(
                        ast.keyword(arg=key, value=ast.Constant(value=val))
                    )

                call_datatable_node.append(
                    ast.Call(
                        func=ast.Name(
                            id=self.registry.get_models_type(self.package_name, typ),
                            ctx=ast.Load(),
                        ),
                        keywords=datatable_keywords,
                    )
                )
            return ast.keyword(
                arg="data_table",
                value=ast.List(elts=call_datatable_node, ctx=ast.Load()),
            )

    def build_step_kwargs(
        self, step_keyword: StepKeyword, stp: GherkinStep
    ) -> list[ast.keyword]:
        """
        Get the step kwargs, e.g. pytest fixtures for the given step.

        :param step_keyword: the step definition keyword.
        :param stp: the step to analyse.

        :return: ast values for the step keyword argument.
        """
        py_kwargs = []
        step_fixtures = self.registry.extract_fixtures(
            self.package_name, step_keyword, stp.text
        )
        for key, _val in step_fixtures.items():
            py_kwargs.append(
                ast.keyword(arg=key, value=ast.Name(id=key, ctx=ast.Load()))
            )

        if stp.doc_string:
            py_kwargs.append(self.parse_doc_string(step_keyword, stp))

        if stp.data_table:
            py_kwargs.append(self.parse_data_table(step_keyword, stp))

        return py_kwargs

    def add_step(
        self,
        stp: GherkinStep,
        stack: list[Any],
        examples: Sequence[GherkinExamples] | None = None,
    ) -> None:
        """
        Appened the given step to the test function.

        :param stp: the step to add.
        :param stack: current compiler stack.
        :param examples: in cases of scenario outline, its associated examples.
        """
        step_keyword = self.get_keyword(stp)
        py_args = self.build_step_args(step_keyword, stp, examples)
        py_kwargs = self.build_step_kwargs(step_keyword, stp)

        call_node = ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="tursu_runner", ctx=ast.Load()),
                attr="run_step",
                ctx=ast.Load(),
            ),
            args=py_args,
            keywords=py_kwargs,
        )

        # Add the call node to the body of the function
        self.step_list.append(ast.Expr(value=call_node, lineno=stp.location.line))

    def to_ast(self) -> ast.FunctionDef:
        """Convert the current state to ast code."""
        return self.funcdef



"""AST helpers at the test module level."""

import ast
from collections.abc import Sequence
from typing import Any

from tursu.domain.model.gherkin import GherkinFeature
from tursu.runtime.registry import Tursu
from tursu.service.ast.astfunction import TestFunctionWriter


class TestModuleWriter:
    """
    Prepare a python test module for the given feature.

    :param feature: the feature that will be compiled to a python module.
    :param registry: tursu registry containing the steps definition.
    :param stack: current stack of the gherkin document traversal.
    """

    def __init__(
        self,
        feature: GherkinFeature,
        registry: Tursu,
        stack: Sequence[Any],
        package_name: str,
    ) -> None:
        self.feature = feature
        self.fixtures: dict[str, type] = {}
        self.registry = registry

        self.package_name = package_name
        self.module_name = stack[0].name
        self.tests_fn: list[ast.stmt] = []

    def append_fixtures(self, fixtures: dict[str, type]) -> None:
        self.fixtures.update(fixtures)

    def import_stmt(self) -> list[ast.stmt]:
        import_mods: list[ast.stmt] = [
            ast.Expr(
                value=ast.Constant(
                    f"{self.feature.name}\n\n{self.feature.description}".strip(),
                    lineno=1,
                )
            ),
            ast.ImportFrom(
                module="typing",
                names=[ast.alias(name="Any", asname=None)],
                level=0,
            ),
            ast.Import(names=[ast.alias(name="pytest", asname=None)]),
            ast.ImportFrom(
                module="tursu.runtime.registry",
                names=[
                    ast.alias(name="Tursu", asname=None),
                ],
                level=0,
            ),
            ast.ImportFrom(
                module="tursu.runtime.runner",
                names=[
                    ast.alias(name="TursuRunner", asname=None),
                ],
                level=0,
            ),
        ]
        for typ, alias in self.registry.get_models_types(self.package_name).items():
            import_mods.append(
                ast.ImportFrom(
                    module=typ.__module__,
                    names=[
                        ast.alias(name=typ.__name__, asname=alias),
                    ],
                    level=0,
                )
            )

        fixtures = self.registry.get_fixtures(self.package_name)

        for key, _typ in self.fixtures.items():
            # register fixture that are declared with their step definition
            if key in fixtures:
                import_mods.append(
                    ast.ImportFrom(
                        module=fixtures[key].__module__,
                        names=[
                            ast.alias(name=fixtures[key].__name__),
                        ],
                        level=0,
                    )
                )
            # we may import the type and type it in the ast function.

        return import_mods

    def append_test(self, fn: TestFunctionWriter) -> None:
        """Append a test function to the module."""
        self.tests_fn.append(fn.to_ast())

    def to_ast(self) -> ast.Module:
        """Convert the current state to ast code."""
        return ast.Module(
            body=self.import_stmt() + self.tests_fn,
            type_ignores=[],
        )



"""Pytest plugin"""

import inspect
import sys
from collections.abc import Iterable
from pathlib import Path
from types import ModuleType
from typing import Any

import pytest

from tursu.domain.model.gherkin import GherkinDocument
from tursu.runtime.registry import Tursu
from tursu.service.compiler import GherkinCompiler

_tursu = Tursu()


@pytest.fixture(scope="session")
def tursu() -> Tursu:
    """
    Fixture used in generated test module to access to the runtime part.

    Tursu generate bytecode for running scenario. The bytecode is simple,
    and is designed to be highly debuggable.

    Tursu object is the registry of the steps definition that can be used
    at runtime to match those definition and run their associated hook.
    """
    return _tursu


def build_pkg(node: Any) -> str:
    parts: list[str] = []
    while node.__class__ is pytest.Package:
        parts.append(node.name)
        node = node.parent
    parts.reverse()
    return ".".join(parts)


class GherkinTestModule(pytest.Module):
    """
    A pytest collector made for gherkin .scenario files.

    While collecting, it generate ast code to run tests using the python runtime.

    :param path: test root directory.
    :param tursu: the tursu registry to register steps.
    :param **kwargs: pytest extra parameters.
    """

    def __init__(
        self,
        path: Path,
        tursu: Tursu,
        parent: pytest.Module | None = None,
        **kwargs: Any,
    ) -> None:
        doc = GherkinDocument.from_file(path)
        self.gherkin_doc = path.name
        assert parent

        self.module_name = build_pkg(parent)
        compiler = GherkinCompiler(doc, tursu, self.module_name)
        self.test_mod = case = compiler.to_module()

        self.test_casefile = path.parent / case.filename
        super().__init__(path=self.test_casefile, parent=parent, **kwargs)
        if (
            self.session.config.getoption("--trace")
            or self.session.config.option.verbose == 3
        ):
            case.write_temporary(path.parent)  # coverage: ignore
            # we preload before updating the path
            self._obj = super()._getobj()  # coverage: ignore

        self._nodeid = self.nodeid.replace(case.filename, path.name)
        self.path = path

    def _getobj(self) -> ModuleType:
        """Convert the scenario to a python module."""
        return self.test_mod.to_python_module()

    def __repr__(self) -> str:
        return f"<GherkinDocument {self.gherkin_doc}>"

    def collect(self) -> Iterable[pytest.Item | pytest.Collector]:
        """Will collect the scenario as an AST generated module."""
        path, self.path = self.path, self.test_casefile  # collect from the ast file
        ret = super().collect()
        self.path = path  # restore the scenario path to have a per path
        return ret


def tursu_collect_file() -> None:
    """
    Used to generate a `pytest_collect_file()` function in a conftest.py file.

    pytest comes with a hook
    [pytest_collect_file](https://docs.pytest.org/en/7.1.x/reference/reference.html#pytest.hookspec.pytest_collect_file)
    used to collect tests that tursu use to generate the tests suite from
    Gherkin files.

    the conftest.py has to be created with in the same directory structure of
    .scenario file.

    Function Tests Directory Structure:

    ```
    tests/functionals/
    │── conftest.py        # Must contains `tursu_collect_file()`
    │── steps.py           # Step definitions for Gherkin scenarios
    │── example.scenario   # Gherkin scenario file
    ```

    If you want to rearrange your scenario in subdirectory, or
    just your steps, it is not a problem, pytest will walk under all directory
    tree. Tursu will collect all steps definifition too.

    ```
    tests/functionals/
    │── conftest.py        # Must contains `tursu_collect_file()`
    │── steps.py           # Common step definitions for Gherkin scenarios
    │── features1/
    │   ├── other_steps.py    # Step definitions for Gherkin scenarios
    │   ├── example.scenario  # Gherkin scenario file
    ```

    ```{important}
    steps definition using the ([@given](#tursu.given),
    [@when](#tursu.when), [@then](#tursu.then)) are globals.

    Says differently, **they are not scoped per sub-directories**.
    ```
    """
    conftest_mod = inspect.getmodule(inspect.stack()[1][0])  # this is conftest.py
    assert conftest_mod

    def pytest_collect_file(  # type: ignore
        parent: pytest.Package, file_path: Path
    ) -> GherkinTestModule | None:
        module_name = conftest_mod.__name__
        parent_name = module_name.rsplit(".", 1)[0]  # Remove the last part
        mod = sys.modules.get(parent_name)
        _tursu.scan(mod)  # load steps before the scenarios

        if file_path.suffix == ".feature":
            doc = GherkinDocument.from_file(file_path)
            ret = GherkinTestModule.from_parent(
                parent, path=file_path, tursu=_tursu, name=doc.name
            )
            return ret

    conftest_mod.pytest_collect_file = pytest_collect_file  # type: ignore



"""Gherkin scenario compiler."""

from collections.abc import Iterator, Sequence
from typing import Any

from tursu.domain.model.gherkin import (
    GherkinBackground,
    GherkinBackgroundEnvelope,
    GherkinDocument,
    GherkinEnvelope,
    GherkinFeature,
    GherkinRuleEnvelope,
    GherkinScenario,
    GherkinScenarioEnvelope,
    GherkinScenarioOutline,
    GherkinStep,
)
from tursu.domain.model.testmod import TestModule
from tursu.runtime.registry import Tursu
from tursu.service.ast.astfunction import TestFunctionWriter
from tursu.service.ast.astmodule import TestModuleWriter


class GherkinIterator:
    """
    Traverse the gherking feature document to emit gherkin objects.
    Keep the stack while traversing.

    :param doc: the document to iterate.
    """

    def __init__(self, doc: GherkinDocument) -> None:
        self.doc = doc
        self.stack: list[Any] = []

    def emit(self) -> Iterator[Any]:
        """Python iterator that emit the stack."""
        self.stack.append(self.doc)
        yield self.stack
        for _ in self.emit_feature(self.doc.feature):
            yield self.stack
        self.stack.pop()

    def emit_feature_from_enveloppe(
        self, enveloppe: Sequence[GherkinEnvelope]
    ) -> Iterator[Any]:
        """
        Helper to traverse Background, Scenario and Rule keywords from the scenario.

        :param enveloppe: Gherkin envelope that wrap the proper object to emit.
        """
        for child in enveloppe:
            match child:
                case GherkinBackgroundEnvelope(background=background):
                    self.stack.append(background)
                    yield self.stack
                    self.stack.pop()
                case GherkinScenarioEnvelope(scenario=scenario):
                    self.stack.append(scenario)
                    yield self.stack
                    for _ in self.emit_scenario(scenario):
                        yield self.stack
                    self.stack.pop()
                case GherkinRuleEnvelope(rule=rule):
                    self.stack.append(rule)
                    yield self.stack
                    for child in self.emit_feature_from_enveloppe(rule.children):
                        yield child
                    self.stack.pop()

    def emit_feature(self, feature: GherkinFeature) -> Iterator[Any]:
        """
        Helper to traverse feature.

        :param feature: Gherkin feature to traverse.
        """
        self.stack.append(feature)
        yield self.stack
        yield from self.emit_feature_from_enveloppe(self.doc.feature.children)
        self.stack.pop()

    def emit_scenario(
        self, scenario: GherkinScenario | GherkinScenarioOutline
    ) -> Iterator[Any]:
        """
        Helper to traverse scenario.

        :param feature: Gherkin scenario or scenario outline to traverse.
        """
        for step in scenario.steps:
            self.stack.append(step)
            yield self.stack
            self.stack.pop()


class GherkinCompiler:
    """
    Gherkin compiler.

    :param doc: the gherkin file to compile.
    :param registry: the tursu registry where steps are already loaded.
    :param package_name: the parent module of the doc.
    """

    def __init__(
        self, doc: GherkinDocument, registry: Tursu, package_name: str
    ) -> None:
        self.emmiter = GherkinIterator(doc)
        self.registry = registry
        self.package_name = package_name

    def to_module(self) -> TestModule:
        """Get the compiled module."""
        module_node = None
        test_function = None
        background_steps: Sequence[GherkinStep] = []

        for stack in self.emmiter.emit():
            el = stack[-1]
            match el:
                case GherkinFeature():
                    assert module_node is None
                    module_node = TestModuleWriter(
                        el, self.registry, stack, self.package_name
                    )

                case GherkinBackground(steps=steps):
                    background_steps = steps

                case GherkinScenario(steps=steps):
                    test_function = TestFunctionWriter(
                        el,
                        self.registry,
                        [*background_steps, *steps],
                        stack,
                        self.package_name,
                    )
                    assert module_node is not None
                    module_node.append_test(test_function)
                    module_node.append_fixtures(test_function.fixtures)
                    if background_steps:
                        for step in background_steps:
                            test_function.add_step(step, stack)
                    for step in steps:
                        test_function.add_step(step, stack)

                case GherkinScenarioOutline(steps=steps, examples=examples):
                    test_function = TestFunctionWriter(
                        el,
                        self.registry,
                        [*background_steps, *steps],
                        stack,
                        self.package_name,
                    )
                    assert module_node is not None
                    module_node.append_test(test_function)
                    module_node.append_fixtures(test_function.fixtures)
                    if background_steps:
                        for step in background_steps:
                            test_function.add_step(step, stack)

                    for step in steps:
                        test_function.add_step(step, stack, examples)

                case _:
                    # print(el)
                    ...

        assert module_node is not None
        return TestModule(module_node.module_name, module_node.to_ast())



"""Test module."""

import ast
import atexit
from pathlib import Path
from types import CodeType, ModuleType


class TestModule:
    """
    AST representation of a scenario.

    :param scenario: the scenario name representated.
    :param module_node: the AST generated code by the compiler of the scenario.
    """

    def __init__(self, scenario: str, module_node: ast.Module) -> None:
        self.scenario = scenario
        self.module_node = module_node

    def __str__(self) -> str:
        return ast.unparse(self.module_node)

    __repr__ = __str__

    @property
    def filename(self) -> str:
        """
        A python test file name used if we wrote the file to the disk.

        Test module are generated in memory, but, while using `--trace`
        tursu write the file to the disk in order to have steps in pdb.

        They are also writen with option `-vvv` in order to have the full traceback.
        """
        return f"test_{self.scenario}.py"

    @property
    def modname(self) -> str:
        """Name of the python module."""
        return self.filename[:-3]

    def compile(self) -> CodeType:
        """Compile ast code to python bytecode."""
        return compile(
            ast.unparse(self.module_node), filename=self.filename, mode="exec"
        )

    def to_python_module(self) -> ModuleType:
        """Generate a python module."""
        mod = ModuleType(self.modname)
        exec(self.compile(), mod.__dict__)
        return mod

    def write_temporary(self, parent: Path) -> None:
        """
        Temporary write the module on the disk. used while tracing to get debug step.
        """
        test_casefile = parent / self.filename
        test_casefile.write_text(str(self), encoding="utf-8")

        def delete_temporary_files() -> None:
            if hasattr(test_casefile, "unlink"):  # coverage: ignore
                test_casefile.unlink(missing_ok=True)  # coverage: ignore

        atexit.register(delete_temporary_files)



"""Registry of step definition."""

import difflib
import importlib
import sys
from collections import defaultdict
from collections.abc import Mapping, Sequence
from inspect import Parameter
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Annotated, Callable, get_args, get_origin

import venusian
from typing_extensions import Any

from tursu.domain.model.steps import Handler, StepDefinition, StepKeyword
from tursu.runtime.pattern_matcher import AbstractPattern

if TYPE_CHECKING:
    from tursu.runtime.runner import TursuRunner

from tursu.runtime.exceptions import Unregistered

VENUSIAN_CATEGORY = "tursu"


def is_init_file(module: ModuleType) -> bool:
    """Check if a module corresponds to an __init__.py file."""
    return (
        hasattr(module, "__file__")
        and module.__file__ is not None
        and Path(module.__file__).name == "__init__.py"
    )


def normalize_module_name(module_name: str) -> str:
    """If the module is an __init__.py, keep its name. Otherwise, drop the last part."""
    module = importlib.import_module(module_name)
    if not is_init_file(module):
        module_name = ".".join(module_name.split(".")[:-1])
    if module_name.endswith(".steps"):
        module_name = module_name[: -len(".steps")]
    return module_name


def _step(
    keyword: StepKeyword, step_pattern: str | AbstractPattern
) -> Callable[[Handler], Handler]:
    def wrapper(wrapped: Handler) -> Handler:
        def callback(scanner: venusian.Scanner, name: str, ob: Handler) -> None:
            if not hasattr(scanner, "registry"):
                return  # coverage: ignore

            step_module = normalize_module_name(ob.__module__)

            scanner.registry.register_step_definition(  # type: ignore
                step_module, keyword, step_pattern, wrapped
            )

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)
        return wrapped

    return wrapper


def given(pattern: str | AbstractPattern) -> Callable[[Handler], Handler]:
    """
    Decorator to listen for the `Given` Gherkin keyword.

    :param pattern: a pattern to extract parameter.
                    Refer to the [step definition documentation](#step-definition)
                    for the syntax.
    :return: the decorate function that have any parameter coming from
             the pattern matcher or pytest fixtures.
    """
    return _step("Given", pattern)


def when(pattern: str | AbstractPattern) -> Callable[[Handler], Handler]:
    """
    Decorator to listen for the `When` gherkin keyword.

    :param pattern: a pattern to extract parameter.
                    Refer to the [step definition documentation](#step-definition)
                    for the syntax.
    :return: the decorate function that have any parameter coming from
             the pattern matcher or pytest fixtures.
    """
    return _step("When", pattern)


def then(pattern: str | AbstractPattern) -> Callable[[Handler], Handler]:
    """
    Decorator to listen for the `Then` gherkin keyword.

    :param pattern: a pattern to extract parameter.
                    Refer to the [step definition documentation](#step-definition)
                    for the syntax.
    :return: the decorate function that have any parameter coming from
             the pattern matcher or pytest fixtures.
    """
    return _step("Then", pattern)


class ModRegistry:
    """
    Registry for a package.

    In a layered tests suite, step definitions are inherits from parent directories,
    so step definitions are saved per directory, or package.
    """

    def __init__(self) -> None:
        self._step_defs: dict[StepKeyword, list[StepDefinition]] = {
            "Given": [],
            "When": [],
            "Then": [],
        }
        self._models_types: dict[type, str] = {}

    @property
    def models_types(self) -> dict[type, str]:
        """
        Registered data types, used in order to build imports on tests.
        The type are aliased during registration to avoid conflict name at import time
        during the ast generation.

        :return: type as key, alias as value.
        """

        return self._models_types

    def append(self, stp: StepKeyword, step: StepDefinition) -> None:
        """
        Append a step definition to the registry.
        """
        self._step_defs[stp].append(step)
        self.register_data_table(step)
        self.register_doc_string(step)

    def get_fixtures(self) -> Mapping[str, type]:
        fixtures: dict[str, type] = {}
        for step_defs in self._step_defs.values():
            for stp in step_defs:
                fixtures.update(stp.fixtures)
        return fixtures

    def register_model(self, parameter: Parameter | None) -> None:
        """
        Register the model in the parameter of a signature for data_table or doc_string.

        :param parameter: the parameter from the signature.
        """
        if parameter and parameter.annotation:
            param_origin = get_origin(parameter.annotation)
            if param_origin is Annotated:
                # we are in a factory
                typ = get_args(parameter.annotation)[-1]
            elif param_origin and issubclass(param_origin, Sequence):
                # we are in a list
                typ = get_args(parameter.annotation)[0]
                item_orig = get_origin(typ)
                if item_orig is not dict:
                    if item_orig is Annotated:
                        # the list has a factory
                        typ = get_args(typ)[-1]
            else:
                # this is a reversed data_table, there should be two column
                typ = parameter.annotation

            if typ.__module__ != "builtins" and typ not in self._models_types:
                self._models_types[typ] = f"{typ.__name__}{len(self._models_types)}"

    def register_data_table(self, step: StepDefinition) -> None:
        """
        This method register the data table as a model.

        :param step: The step containing a data_table parameter.
        """
        self.register_model(step.pattern.signature.parameters.get("data_table"))

    def register_doc_string(self, step: StepDefinition) -> None:
        """
        This method register the doc string as a model.

        :param step: The step containing a doc_string parameter.
        """
        self.register_model(step.pattern.signature.parameters.get("doc_string"))

    def get_best_matches(
        self,
        text: str,
        n: int = 5,
        cutoff: float = 0.3,
    ) -> Sequence[tuple[float, str]]:
        """
        Return the gherkin steps from the registry that look like the given text.
        This method is called if no step definition matches to build a proper hint
        for the user.

        :param text: text to match the definition.
        """
        possibilities = [
            *[f"Given {stp.pattern.pattern}" for stp in self._step_defs["Given"]],
            *[f"When {stp.pattern.pattern}" for stp in self._step_defs["When"]],
            *[f"Then {stp.pattern.pattern}" for stp in self._step_defs["Then"]],
        ]
        matches = difflib.get_close_matches(text, possibilities, n=n, cutoff=cutoff)

        scored_matches = [
            (difflib.SequenceMatcher(None, text, match).ratio(), match)
            for match in matches
        ]
        return scored_matches

    def get_step(self, keyword: StepKeyword, text: str) -> StepDefinition | None:
        """
        Get the first registered step that match the text.

        :param keyword: gherkin keyword for the definition.
        :param text: text to match the definition.
        :return: the register step if exists otherwise None.
        """

        step_defs = self._step_defs[keyword]
        for stp in step_defs:
            if stp.pattern.match(text):
                return stp
        return None


class Registry:
    """
    Layered step definitions registry.

    A facade for {class}`ModRegistry` to have step definitions in a tree.
    """

    def __init__(self) -> None:
        self._step_defs: dict[str, ModRegistry] = defaultdict(ModRegistry)

    def append(
        self, module_name: str, keyword: StepKeyword, step: StepDefinition
    ) -> None:
        """
        Append the step definition to the module name registry.

        :param module_name: the name of the module of the step is registered.
        :param keyword: gherkin keyword for the definition.
        :param step: step definition to append.
        """
        self._step_defs[module_name].append(keyword, step)

    def get_fixtures(self, module_name: str) -> Mapping[str, type]:
        """
        Return the list of fixtures that has to be imported at the module.

        :return: a mapping where the key is the alias name and value is the
            type to import.
        """
        fixtures: dict[str, type] = {}

        parts = module_name.split(".")
        module_name = parts.pop(0)
        while True:
            if module_name in self._step_defs:
                fixtures.update(self._step_defs[module_name].get_fixtures())
            if parts:
                module_name = f"{module_name}.{parts.pop(0)}"
            else:
                break
        return fixtures

    def get_models_types(self, module_name: str) -> dict[type, str]:
        """
        Registered data types, used in order to build imports on tests.
        The type are aliased during registration to avoid conflict name at import time
        during the ast generation.

        :param module_name: the name of the module the step is retrieve,
            and lookup from the current module to its ancestors,
            retrieving the first that match.
        :return: type as key, alias as value.
        """
        model_types: dict[type, str] = {}

        parts = module_name.split(".")
        module_name = parts.pop(0)
        while True:
            if module_name in self._step_defs:
                model_types.update(self._step_defs[module_name].models_types)
            if parts:
                module_name = f"{module_name}.{parts.pop(0)}"
            else:
                break
        return model_types

    def get_step(
        self, module_name: str, keyword: StepKeyword, text: str
    ) -> StepDefinition | None:
        """
        Get the first registered step that match the text.

        :param module_name: the name of the module the step is retrieve,
            and lookup from the current module to its ancestors,
            retrieving the first that match.
        :param keyword: gherkin keyword for the definition.
        :param text: text to match the definition.
        :return: the register step if exists otherwise None.
        """
        parts = module_name.split(".")
        while parts:
            mod_path = ".".join(parts)
            if mod_path in self._step_defs:
                if handle := self._step_defs[mod_path].get_step(keyword, text):
                    return handle
            parts.pop()
        return None

    def get_matched_step(
        self,
        module_name: str,
        keyword: StepKeyword,
        text: str,
        fixtures: Mapping[str, Any],
    ) -> tuple[StepDefinition | None, Mapping[str, Any]]:
        """
        Get the first registered step that match the text.

        :param module_name: the name of the module the step is retrieve,
            and lookup from the current module to its ancestors,
            retrieving the first that match.
        :param keyword: gherkin keyword for the definition.
        :param text: text to match the definition.
        :return: the register step if exists otherwise None with its associated step
            definition parameters.
        """
        step_def = self.get_step(module_name, keyword, text)
        if step_def:
            matches = step_def.pattern.get_matches(text, fixtures)
            return step_def, matches or {}
        return None, {}

    def get_best_matches(
        self,
        module_name: str,
        text: str,
    ) -> list[str]:
        """
        Get the first registered step that match the text.

        :param module_name: use the module and all its ancestors to retrieve a match.
        :param text: text to match the definition.
        :return: the list of steps that could be usefull.
        """
        parts = module_name.split(".")
        matches: list[tuple[float, str]] = []
        while parts:
            mod_path = ".".join(parts)
            if mod_path in self._step_defs:
                matches += self._step_defs[mod_path].get_best_matches(text)
            parts.pop()
        matches = list(set(matches))
        matches.sort(reverse=True)
        sure_threshold = 0.9
        while sure_threshold >= 0.4:
            match_text = [match[1] for match in matches if match[0] > sure_threshold]
            if match_text:
                return match_text
            sure_threshold -= 0.1
        return []


class Tursu:
    """Store all the handlers for gherkin action."""

    DATA_TABLE_EMPTY_CELL = ""
    """
    This value is used only in case of data_table types usage.
    If the table contains this value, then, it is ommited by the constructor in order
    to let the type default value works.

    In case of list[dict[str,str]], then this is ignored, empty cells exists with
    an empty string value.
    """

    def __init__(self) -> None:
        self.scanned: set[ModuleType] = set()
        self._registry = Registry()

    def get_fixtures(self, module_name: str) -> Mapping[str, type]:
        """
        Return the list of fixtures that has to be imported at the module.

        :return: a mapping where the key is the alias name and value is the
            type to import.
        """
        return self._registry.get_fixtures(module_name)

    def register_step_definition(
        self,
        module_name: str,
        keyword: StepKeyword,
        pattern: str | AbstractPattern,
        handler: Handler,
    ) -> None:
        """
        Register a step handler for a step definition.

        This method is the primitive for [@given](#tursu.given),
        [@when](#tursu.when) and [@then](#tursu.then) decorators.

        :param module_name: the name of the module of the step is registered.
        :param keyword: gherkin keyword for the definition.
        :param pattern: pattern to match the definition.
        :param handler: function called when a step in a scenario match the pattern.
        """
        step = StepDefinition(pattern, handler)
        self._registry.append(module_name, keyword, step)

    def get_step(
        self, module_name: str, keyword: StepKeyword, text: str
    ) -> StepDefinition | None:
        """
        Get the first registered step that match the text.

        :param module_name: the name of the current module.
        :param keyword: gherkin keyword for the definition.
        :param text: text to match the definition.
        :return: the register step if exists otherwise None.
        """
        return self._registry.get_step(module_name, keyword, text)

    def get_models_types(self, module_name: str) -> dict[type, str]:
        """
        Registered data types, used in order to build imports on tests.
        The type are aliased during registration to avoid conflict name at import time
        during the ast generation.

        :param module_name: the name of the module the step is retrieve,
            and lookup from the current module to its ancestors,
            retrieving the first that match.
        :return: type as key, alias as value.
        """
        return self._registry.get_models_types(module_name)

    def get_models_type(self, module_name: str, typ: type[Any]) -> str:
        """
        Get the alias name of a given type used to build the model type
        during AST compilation.

        :param module_name: the name of the module the step is retrieve,
            and lookup from the current module to its ancestors,
            retrieving the first that match.
        :param typ: the type of the model.
        :return: alias of the type.
        """
        return self._registry.get_models_types(module_name)[typ]

    def get_best_matches(self, module_name: str, text: str) -> list[str]:
        return self._registry.get_best_matches(module_name, text)

    def extract_fixtures(
        self, module_name: str, keyword: StepKeyword, text: str, **kwargs: Any
    ) -> Mapping[str, Any]:
        """
        Extract fixture for a step from the given pytest fixtures of the test function.

        :param module_name: module name where the step has to be found.
        :param keyword: gherkin step to match.
        :param text: text to match the definition.
        :param kwargs: the fixtures pytest fixtures from the test function.
        :return: the fixtures for the step handler.
        """
        step = self._registry.get_step(module_name, keyword, text)
        if step is None:
            raise Unregistered(module_name, self, keyword, text)
        return step.pattern.extract_fixtures(text) or {}

    def run_step(
        self,
        tursu_runner: "TursuRunner",
        keyword: StepKeyword,
        text: str,
        **kwargs: Any,
    ) -> None:
        """
        Run the step that match the parameter and emit information to the runner.

        :param tursu_runner: the fixtures pytest fixtures from the test function.
        :param keyword: gherkin step to match.
        :param text: text to match the definition.
        :param kwargs: the fixtures pytest fixtures from the test function.
        """
        handler, matches = self._registry.get_matched_step(
            tursu_runner.module_name, keyword, text, kwargs
        )
        if handler:
            tursu_runner.emit_running(keyword, handler, matches)
            try:
                handler(**matches)
            except Exception:
                tursu_runner.emit_error(keyword, handler, matches)
                raise
            else:
                tursu_runner.emit_success(keyword, handler, matches)
        else:
            tursu_runner.emit_error(
                keyword, StepDefinition(text, lambda: None), {}, unregistered=True
            )
            raise Unregistered(tursu_runner.module_name, self, keyword, text)

    def scan(self, mod: ModuleType | None = None) -> "Tursu":
        """
        Scan the module (or modules) containing steps.

        :param mod: optionally module to import, usually the function caller module.
        :return: the current tursu registry for multiple scan purpose.
        """
        if mod is None:
            import inspect

            mod = inspect.getmodule(inspect.stack()[1][0])
            assert mod
            module_name = mod.__name__
            if "." in module_name:  # Check if it's a submodule
                parent_name = module_name.rsplit(".", 1)[0]  # Remove the last part
                mod = sys.modules.get(parent_name)

        assert mod
        if mod not in self.scanned:
            self.scanned.add(mod)
            scanner = venusian.Scanner(registry=self)
            scanner.scan(mod, categories=[VENUSIAN_CATEGORY])
        return self



"""Run a gherkin scenario."""

import logging
import re
import time
from collections.abc import Mapping
from types import TracebackType
from typing import Self

import pytest
from typing_extensions import Any

from tursu.domain.model.steps import StepDefinition, StepKeyword
from tursu.runtime.registry import Tursu

# Set up the logger
logger = logging.getLogger("tursu")
logger.setLevel(logging.DEBUG)


# ANSI color codes
GREEN = "\033[92m"
GREY = "\033[90m"
ORANGE = "\033[93m"
RED = "\033[91m"
CYAN = "\033[36m"

RESET = "\033[0m"  # Reset color
UP = "\033[F"  # Cursor Previous Line
EL = "\033[K"  # Erase in line


class ScenarioFailed(Exception):
    """Scenario failure error."""


class TursuRunner:
    """
    Run the scenario in a context manager.

    :param request: the pytest request fixture.
    :param capsys: the pytest capsys fixture.
    :param tursu: the tursu registry.
    :param scenario: the stack list of gherkin sentence run for display purpose.
    """

    IGNORE_TIMING_MS = 200
    OK_TIMING_MS = 700
    WARN_TIMING_MS = 2100

    def __init__(
        self,
        request: pytest.FixtureRequest,
        capsys: pytest.CaptureFixture[str],
        tursu: Tursu,
        scenario: list[str],
    ) -> None:
        self.module_name = request.node.parent.module_name
        self.name = request.node.nodeid
        self.verbose = request.config.option.verbose
        self.tursu = tursu
        self.capsys = capsys
        self.runned: list[str] = []
        self.scenario = scenario
        self.start_time = time.perf_counter()
        self.end_time = time.perf_counter()

        if self.verbose:
            self.log("", replace_previous_line=True)
            for step in self.scenario:
                self.log(step)

    def remove_ansi_escape_sequences(self, text: str) -> str:
        """
        Sanitize text of terminal decoration.

        :param text: the text to cleanup.
        """
        return re.sub(r"\x1b\[[0-9;]*[a-zA-Z]", "", text)

    def fancy(self) -> str:
        """Terminal fancy representation of the current state of the runner."""
        lines: list[str] = self.runned or ["🔥 no step runned"]
        lines = self.scenario + lines
        line_lengthes = [len(self.remove_ansi_escape_sequences(line)) for line in lines]
        max_line_length = max(line_lengthes)

        # Create the border based on the longest line
        top_border = f"{RED}┌" + "─" * (max_line_length + 3) + f"┐{RESET}"
        bottom_border = f"{RED}└" + "─" * (max_line_length + 3) + f"┘{RESET}"

        middle_lines = []
        sep = f"{RED}│{RESET}"
        for line, length in zip(lines, line_lengthes):
            middle_lines.append(
                f"{sep} {line + ' ' * (max_line_length - length)} {sep}"
            )

        middle_lines_str = "\n".join(middle_lines)
        return f"\n{top_border}\n{middle_lines_str}\n{bottom_border}\n"

    def log(
        self, text: str, replace_previous_line: bool = False, end: str = "\n"
    ) -> None:
        """Helper method to log line."""
        if self.verbose:  # coverage: ignore
            with self.capsys.disabled():  # coverage: ignore
                if replace_previous_line and self.verbose == 1:  # coverage: ignore
                    print(UP, end="")  # coverage: ignore
                print(f"{text}{EL}", end=end)  # coverage: ignore

    def run_step(
        self,
        step: StepKeyword,
        text: str,
        **kwargs: Any,
    ) -> None:
        """
        Will run the given step using the tursu registry, raised an error if its fail.

        :param step: gherkin keyword.
        :param text: text that should match a step definition.

        :raises ScenarioFailed: if the step did not run properly.
        """
        try:
            self.tursu.run_step(self, step, text, **kwargs)
        except Exception as exc:
            raise ScenarioFailed(self.fancy()) from exc

    def format_example_step(self, text: str, **kwargs: Any) -> str:
        """
        Format the scenario outline with args that comes from the parametrized mark.

        :param text: gherkin step from scenario file.
        :param **kwargs: example line for the Examples of the scenario outline.
        """
        for key, val in kwargs.items():
            text = text.replace(f"<{key}>", val)
        return text

    def emit_running(
        self,
        keyword: StepKeyword,
        step: StepDefinition,
        matches: Mapping[str, Any],
    ) -> None:
        """
        Update state when a step is marked as running.

        :param keyword: gherkin step keyword.
        :param step: matched step for the tursu registry.
        :param matches: parameters that match for highlighting purpose.
        """
        text = f"{GREY}⏳ {keyword} {step.highlight(matches, CYAN, GREY)}{RESET}"
        self.runned.append(text)
        self.log(text)
        self.start_time = time.perf_counter()

    def emit_error(
        self,
        keyword: StepKeyword,
        step: StepDefinition,
        matches: Mapping[str, Any],
        *,
        unregistered: bool = False,
    ) -> None:
        """
        Update state when a step is marked as error.

        :param keyword: gherkin step keyword.
        :param step: matched step for the tursu registry.
        :param matches: parameters that match for highlighting purpose.
        """
        self.end_time = time.perf_counter()
        elapsed_ms = (self.end_time - self.start_time) * 1000

        timelog = (
            f" {RED}[{elapsed_ms:.2f}ms]"
            if elapsed_ms > self.IGNORE_TIMING_MS or self.verbose > 1
            else ""
        )

        steplog = step.highlight(matches, CYAN, RED)
        text = f"{RED}❌ {keyword}{RESET} {steplog}{timelog}{RESET}"
        if not unregistered:
            self.runned.pop()
        self.runned.append(text)
        self.log(text, True)
        self.log("-" * (len(self.name) + 2), end="")

    def emit_success(
        self, keyword: StepKeyword, step: StepDefinition, matches: Mapping[str, Any]
    ) -> None:
        """
        Update state when a step is marked as success.

        :param keyword: gherkin step keyword.
        :param step: matched step for the tursu registry.
        :param matches: parameters that match for highlighting purpose.
        """

        self.end_time = time.perf_counter()
        elapsed_ms = (self.end_time - self.start_time) * 1000

        if elapsed_ms < self.OK_TIMING_MS:
            color = GREEN
        elif elapsed_ms < self.WARN_TIMING_MS:
            color = ORANGE
        else:
            color = RED

        timelog = (
            f" {color}[{elapsed_ms:.2f}ms]"
            if elapsed_ms > self.IGNORE_TIMING_MS or self.verbose > 1
            else ""
        )

        steplog = step.highlight(matches, CYAN, GREEN)
        text = f"{GREEN}✅ {keyword} {steplog}{timelog}{RESET}"
        self.runned.pop()
        self.runned.append(text)
        self.log(text, True)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        self.log(" " * (len(self.name) + 2), end="")



"""Step definition hooks."""

import inspect
import sys
from collections.abc import Mapping
from typing import Any, Callable, Literal

from tursu.runtime.pattern_matcher import (
    AbstractPattern,
    AbstractPatternMatcher,
    DefaultPatternMatcher,
)

StepKeyword = Literal["Given", "When", "Then"]
"""Gherkin keywords that can be mapped to step definitions."""

Handler = Callable[..., None]
"""
The hook handler is a decorated function that have any parameters
but can't return anything.

The decorated method parameters comes from the pattern matcher first
and fallback to pytest fixtures.
"""


def discover_fixtures(hook: Callable[..., None]) -> dict[str, type]:
    """
    Get all the fixtures that have been declared in the hook module
    from the signature of the hook.
    """
    signature = inspect.signature(hook)
    module = sys.modules[hook.__module__]
    fixtures: dict[str, type] = {}
    for key in signature.parameters:
        if hasattr(module, key):
            fixtures[key] = getattr(module, key)
    return fixtures


class StepDefinition:
    """
    Step definition.

    :param pattern: pattern matcher for the step.
    :param hook: The decorated method.
    """

    def __init__(self, pattern: str | AbstractPattern, hook: Handler):
        matcher: type[AbstractPatternMatcher]
        if isinstance(pattern, str):
            matcher = DefaultPatternMatcher
        else:
            matcher = pattern.get_matcher()
            pattern = pattern.pattern

        self.pattern = matcher(pattern, inspect.signature(hook))
        self.hook = hook
        self.fixtures: Mapping[str, type] = discover_fixtures(hook)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, StepDefinition):
            return False
        return self.pattern == other.pattern and self.hook == other.hook

    def __repr__(self) -> str:
        return f'StepDefinition("{self.pattern}", {self.hook.__qualname__})'

    def __call__(self, **kwargs: Any) -> None:
        """Will call the hook with the given parameter."""
        self.hook(**kwargs)

    def highlight(
        self,
        matches: Mapping[str, Any],
        color: str = "\033[36m",
        reset: str = "\033[0m",
    ) -> str:
        """Highlith representation of a step that has matched for the terminal."""
        return self.pattern.hightlight(matches, color, reset)



"""Runtime exception"""

import textwrap
from typing import TYPE_CHECKING

from tursu.domain.model.steps import StepKeyword

if TYPE_CHECKING:
    from .registry import Tursu

TEMPLATE_WITH_MATCHED_STEPS = """
Unregistered step:

    {step} {text}

Maybe you look for:

    {registered_list_str}

Otherwise, to register this new step:
{create_step}
"""

TEMPLATE_WITHOUT_MATCHED_STEPS = """
Unregistered step:

    {step} {text}

To register this new step:
{create_step}
"""


class Unregistered(RuntimeError):
    """
    Raised when no step definition are found from a gherkin step.

    :param module_name: the test module where the step definition is not registered.
    :param registry: the tursu registry.
    :param step: Keyworkd of the step.
    :param text: the text that did not match any step definition.
    """

    def __init__(
        self, module_name: str, registry: "Tursu", step: StepKeyword, text: str
    ):
        registered_list: list[str] = registry.get_best_matches(module_name, text)

        registered_list_str = "\n    ".join(registered_list)
        safe_text = text.replace('"', '\\"')
        create_step = textwrap.indent(
            textwrap.dedent(
                f"""
                @{step.lower()}("{safe_text}")
                def step_definition(): ...
                """
            ),
            prefix="    ",
        )

        template = (
            TEMPLATE_WITH_MATCHED_STEPS
            if registered_list
            else TEMPLATE_WITHOUT_MATCHED_STEPS
        )
        super().__init__(
            template.format(
                step=step,
                text=text,
                registered_list_str=registered_list_str,
                create_step=create_step,
            )
        )



"""
Gherkin feature files as DOM objects.

Turşu use the gherkin-official package to parse file, with a pydantic
layer in order to works with typed object instead of typed dict.

It also clean up keywords and add representation for the output.
"""

import ast
import json
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Annotated, Any, Literal

from gherkin import Parser
from pydantic import BaseModel, Field, model_validator
from pydantic.functional_validators import BeforeValidator


def sanitize(value: Any) -> str:
    return value.strip().title() if isinstance(value, str) else value


GherkinKeyword = Annotated[
    Literal[
        "Feature",
        "Scenario",
        "Scenario Outline",
        "Examples",
        "Background",
        "Rule",
        "Given",
        "When",
        "Then",
        "And",
        "But",
    ],
    BeforeValidator(sanitize),
]

GherkinScenarioKeyword = Annotated[Literal["Scenario"], BeforeValidator(sanitize)]
GherkinScenarioOutlineKeyword = Annotated[
    Literal["Scenario Outline"], BeforeValidator(sanitize)
]

StrippedWhitespace = Annotated[
    str, BeforeValidator(lambda value: value.strip() if value else value)
]


class GherkinLocation(BaseModel):
    line: int
    column: int | None = Field(default=None)


class GherkinComment(BaseModel):
    location: GherkinLocation
    text: str


class GherkinTag(BaseModel):
    id: str
    location: GherkinLocation
    name: Annotated[
        str,
        BeforeValidator(lambda value: value.strip().lstrip("@") if value else value),
    ]


class GherkinCell(BaseModel):
    location: GherkinLocation
    value: str


class GherkinTableRow(BaseModel):
    id: str
    location: GherkinLocation
    cells: Sequence[GherkinCell]


class GherkinDataTable(BaseModel):
    location: GherkinLocation
    rows: list[GherkinTableRow]


class GherkinDocString(BaseModel):
    location: GherkinLocation
    content: str | Mapping[str, Any] | Sequence[Any]
    delimiter: str
    media_type: str | None = Field(default=None, alias="mediaType")

    @model_validator(mode="after")
    def check_passwords_match(self) -> "GherkinDocString":
        if not isinstance(self.content, str):
            return self
        match self.media_type:
            case "json":
                self.content = json.loads(self.content)
            case "python":
                self.content = ast.literal_eval(self.content)
            case _:
                ...
        return self


class GherkinStep(BaseModel):
    id: str
    location: GherkinLocation
    keyword: GherkinKeyword
    text: str
    keyword_type: str = Field(alias="keywordType")
    data_table: GherkinDataTable | None = Field(default=None, alias="dataTable")
    doc_string: GherkinDocString | None = Field(default=None, alias="docString")

    def __repr__(self) -> str:
        return f"{self.keyword} {self.text}"


class GherkinBackground(BaseModel):
    id: str
    location: GherkinLocation
    keyword: GherkinKeyword
    name: StrippedWhitespace
    description: StrippedWhitespace
    steps: Sequence[GherkinStep]

    def __repr__(self) -> str:
        return f"Background: {self.name}"


class GherkinExamples(BaseModel):
    id: str
    location: GherkinLocation
    tags: Sequence[GherkinTag]
    keyword: GherkinKeyword
    name: StrippedWhitespace
    description: StrippedWhitespace
    table_header: GherkinTableRow = Field(alias="tableHeader")
    table_body: Sequence[GherkinTableRow] = Field(alias="tableBody")


class GherkinScenario(BaseModel):
    id: str
    location: GherkinLocation
    tags: Sequence[GherkinTag]
    keyword: GherkinScenarioKeyword
    name: StrippedWhitespace
    description: StrippedWhitespace
    steps: Sequence[GherkinStep]

    def __repr__(self) -> str:
        return f"🎬 Scenario: {self.name}"


class GherkinScenarioOutline(BaseModel):
    id: str
    location: GherkinLocation
    tags: Sequence[GherkinTag]
    keyword: GherkinScenarioOutlineKeyword
    name: StrippedWhitespace
    description: StrippedWhitespace
    steps: Sequence[GherkinStep]
    examples: Sequence[GherkinExamples]

    def __repr__(self) -> str:
        return f"🎬 Scenario Outline: {self.name}"


class GherkinBackgroundEnvelope(BaseModel):
    background: GherkinBackground


class GherkinScenarioEnvelope(BaseModel):
    scenario: GherkinScenario | GherkinScenarioOutline


class GherkinRuleEnvelope(BaseModel):
    rule: "GherkinRule"


GherkinEnvelope = (
    GherkinBackgroundEnvelope | GherkinScenarioEnvelope | GherkinRuleEnvelope
)


class GherkinRule(BaseModel):
    id: str
    location: GherkinLocation
    tags: Sequence[GherkinTag]
    keyword: GherkinKeyword
    name: StrippedWhitespace
    description: StrippedWhitespace
    children: Sequence[GherkinEnvelope]

    def __repr__(self) -> str:
        return f"🔹 Rule: {self.name}"


class GherkinFeature(BaseModel):
    location: GherkinLocation
    tags: Sequence[GherkinTag]
    language: str
    keyword: GherkinKeyword
    name: StrippedWhitespace
    description: StrippedWhitespace
    children: Sequence[GherkinEnvelope]

    def __repr__(self) -> str:
        return f"🥒 Feature: {self.name}"


class GherkinDocument(BaseModel):
    name: StrippedWhitespace
    filepath: Path
    feature: GherkinFeature
    comments: Sequence[GherkinComment]

    @classmethod
    def from_file(cls, file: Path) -> "GherkinDocument":
        """
        Parse .scenario file from the given path and build the gherkin document.

        :param file: path on disk to the scenario file.
        """
        official_doc = Parser().parse(file.read_text())
        return GherkinDocument(
            name=file.name[: -len(".feature")],
            filepath=file,
            **official_doc,  # type: ignore
        )

    def __repr__(self) -> str:
        return f"📄 Document: {self.name}.feature"


class Stack(BaseModel):
    value: list[BaseModel]



"""
Pattern matcher module is used to make the link between gherkin step from a scenario

and step from the registry, registered via the [@given](#tursu.given),
[@when](#tursu.when) and [@then](#tursu.then) decorators.
"""

import abc
import re
from collections.abc import Mapping
from datetime import date, datetime
from enum import Enum
from inspect import Signature
from typing import Any, Literal, get_args, get_origin


class PatternError(RuntimeError):
    """
    Raised if a pattern is invalid.

    This exception happens during the log of the [step registry](#tursu.runtime.registry.Tursu).
    """


def cast_to_annotation(
    value: str, annotation: type[int | float | bool | str | date | datetime | Enum]
) -> int | float | bool | str | date | datetime | Enum:
    """
    Safely casts a string to the given annotation.

    :param value: The parameter to be cast.
    :param annotation: the constructor.

    :return: The casted value if successful, otherwise raises a ValueError.
    """

    # Define safe standard library types
    safe_types = (int, float, bool, str, date, datetime)
    if annotation in safe_types:
        # Handle special case for bool
        if annotation is bool:
            true_vals = {"true", "1", "yes", "on"}
            false_vals = {"false", "0", "no", "off"}
            lower_val = value.lower()
            if lower_val in true_vals:
                return True
            elif lower_val in false_vals:
                return False
            else:
                raise ValueError(
                    f"Cannot cast '{value}' to bool: "
                    f"use one of {(', ').join(sorted([*true_vals, *false_vals]))}"
                )

        if annotation is date:
            try:
                return date.fromisoformat(value)
            except ValueError:
                raise ValueError(
                    f"Cannot cast '{value}' to date: use iso format"
                ) from None

        if annotation is datetime:
            try:
                return datetime.fromisoformat(value)
            except ValueError:
                raise ValueError(
                    f"Cannot cast '{value}' to datetime: use iso format"
                ) from None

        try:
            # the type can't be a date or a datetime anymore
            return annotation(value)  # type: ignore
        except (ValueError, TypeError) as exc:
            raise ValueError(f"Cannot cast '{value}' to {annotation}: {exc}") from exc

    if get_origin(annotation) is Literal:
        if value in get_args(annotation):
            return value
        raise ValueError(
            f"Value '{value}' is not a valid Literal: {get_args(annotation)}"
        )

    if isinstance(annotation, type) and issubclass(annotation, Enum):  # type: ignore
        try:
            return annotation[value]
        except KeyError:
            raise ValueError(
                f"Cannot cast '{value}' to Enum {annotation.__name__}"
            ) from None

    raise TypeError(f"Unsafe or unsupported type: {annotation}")


class AbstractPatternMatcher(abc.ABC):
    """
    Base class to that implement the maghin pattern.
    """

    pattern: str
    """The string reprentation of the pattern to match."""
    signature: Signature
    """
    The decorated function signature used for introspecting fixtures and type hinting.
    """

    def __init__(self, pattern: str, signature: Signature) -> None:
        """
        :param pattern: Step decorator text.
        :param signature: Signature of the decorated method.
        """
        self.pattern = pattern
        self.signature = signature

    def __eq__(self, other: Any) -> bool:
        if self.__class__ != other.__class__:
            return False
        return self.pattern == self.pattern

    def __str__(self) -> str:
        return self.pattern

    def __repr__(self) -> str:
        return f'"{self.pattern}"'

    @abc.abstractmethod
    def match(self, text: str) -> bool:
        """Return true if its a match"""

    @abc.abstractmethod
    def get_matches(
        self, text: str, fixtures: Mapping[str, Any]
    ) -> Mapping[str, Any] | None:
        """
        Used to know if the gherkin step from a scenario has matched the step,

        and return the associated paramters.

        :param text: the text to match
        :param fixtures: the fixtures provided trom the test function, that have
                         to be passed to fill the parameters of the function
                         not provided by the matcher.
        """

    @abc.abstractmethod
    def extract_fixtures(self, text: str) -> Mapping[str, Any] | None:
        """Get the fixtures list to use from the text."""

    @abc.abstractmethod
    def hightlight(
        self,
        matches: Mapping[str, Any],
        color: str = "\033[36m",
        reset: str = "\033[0m",
    ) -> str:
        """
        Return a text representation of the step with the matched text highlighted
        for the terminal.

        :param matched: the list of matched step to consume.
        :return: the highlighted version.
        """


class AbstractPattern(abc.ABC):
    """
    Identifier of a pattern in a step.

    Technically its an
    [AbstractPatternMatcher](#tursu.runtime.pattern_matcher.AbstractPatternMatcher) factory.
    """

    pattern: str
    """The text representation of the pattern in the step."""

    def __init__(self, pattern: str) -> None:
        self.pattern = pattern

    @classmethod
    @abc.abstractmethod
    def get_matcher(cls) -> type[AbstractPatternMatcher]:
        """Return the appropriate matcher for the step."""


class RegexBasePattern(AbstractPatternMatcher):
    """Base class for pattern matcher that consume a regex."""

    re_pattern: re.Pattern[str]
    """The compiled version of the pattern."""

    def match(self, text: str) -> bool:
        matches = self.re_pattern.match(text)
        return bool(matches)

    def get_matches(
        self, text: str, fixtures: Mapping[str, Any]
    ) -> Mapping[str, Any] | None:
        """
        Used to know if the gherkin step from a scenario has matched the step,

        and return the associated paramters.

        :param text: the text to match
        :param fixtures: the fixtures provided trom the test function, that have
                         to be passed to fill the parameters of the function
                         not provided by the matcher.
        """

        matches = self.re_pattern.match(text)
        if matches:
            res = {}
            matchdict = matches.groupdict()
            for key, val in self.signature.parameters.items():
                if key in matchdict:
                    # transform the annotation to call the constructror with the value
                    typ = self.signature.parameters[key].annotation
                    res[key] = cast_to_annotation(matchdict[key], typ)
                elif key in fixtures:
                    res[key] = fixtures[key]
                elif val.default and val.default != val.empty:
                    res[key] = val.default
            return res

        return None

    def extract_fixtures(self, text: str) -> Mapping[str, Any] | None:
        """
        Get the fixtures list to use from the text.

        :param text: the text from a gherkin step in a scenario.
        """
        matches = self.re_pattern.match(text)
        if matches:
            res = {}
            matchdict = matches.groupdict()
            for key, val in self.signature.parameters.items():
                if key in matchdict:
                    continue
                elif key in ("doc_string", "data_table"):
                    continue
                if val.default != val.empty:
                    continue
                res[key] = val.annotation
            return res

        return None


class DefaultPatternMatcher(RegexBasePattern):
    """
    The pattern matcher that look like the python format.
    """

    def __init__(self, pattern: str, signature: Signature) -> None:
        """
        :param pattern: Step decorator text.
        :param signature: Signature of the decorated method.
        """
        super().__init__(pattern, signature)
        re_pattern = pattern
        for key, val in signature.parameters.items():
            match val.annotation:
                case type() if val.annotation is int:
                    re_pattern = re_pattern.replace(f"{{{key}}}", rf"(?P<{key}>\d+)")
                case _:
                    # if enclosed by double quote, use double quote as escaper
                    # not a gherkin spec.
                    re_pattern = re_pattern.replace(
                        f'"{{{key}}}"', rf'"(?P<{key}>[^"]+)"'
                    )
                    # otherwise, match one word
                    re_pattern = re_pattern.replace(f"{{{key}}}", rf"(?P<{key}>[^\s]+)")
        self.re_pattern = re.compile(f"^{re_pattern}$")

    def hightlight(
        self,
        matches: Mapping[str, Any],
        color: str = "\033[36m",
        reset: str = "\033[0m",
    ) -> str:
        """
        Return a text representation of the step with the matched text highlighted
        for the terminal.

        :param matched: the list of matched step to consume.
        :return: the highlighted version.
        """
        colored_matches = {
            key: f"{color}{value}{reset}" for key, value in matches.items()
        }
        return self.pattern.format(**colored_matches)


class RegExPatternMatcher(RegexBasePattern):
    """
    The pattern matcher that look like the python regular expression.
    """

    def __init__(self, pattern: str, signature: Signature) -> None:
        """
        :param pattern: Step decorator text that contain the regex.
        :param signature: Signature of the decorated method.
        """
        super().__init__(pattern, signature)
        try:
            self.re_pattern = re.compile(f"^{pattern}$")
        except re.PatternError as exc:
            raise PatternError(f"Can't compile to regex: {pattern}: {exc}") from exc

    def hightlight(
        self,
        matches: Mapping[str, Any],
        color: str = "\033[36m",
        reset: str = "\033[0m",
    ) -> str:
        """
        Return a text representation of the step with the matched text highlighted
        for the terminal.

        :param matched: the list of matched step to consume.
        :return: the highlighted version.
        """
        colored_matches = {
            key: f"{color}{value}{reset}" for key, value in matches.items()
        }
        ret = self.pattern
        for key, val in colored_matches.items():
            ret = re.sub(rf"\(\?P<{key}>[^\)]+\)", val, ret)
        return ret


class RegEx(AbstractPattern):
    """Marker for regex in a step decorator."""

    @classmethod
    def get_matcher(cls) -> type[AbstractPatternMatcher]:
        """Return the appropriate matcher for the step."""
        return RegExPatternMatcher

    def __repr__(self) -> str:
        return f'r"{self.pattern}"'
